/*
 * DbObject.java
 * 2002-04-11  Knabe  Splitted into interface DbObject and class DbObjectImpl
 * 2002-04-11  Knabe  All operation names except getOid() begin with prefix db
 * Created on 28. Februar 2001, 15:31
 */

package bank3tier.db;

/**Base interface for all persistent objects. This interface is purely generic.
  All application specific code (especially table creation) is collected
  in a subclass of {@link DbApplication}.
  Each object is identified by an OID. The OID (Object IDentification) is a
  number unique for each object in our application.
  All operation names except the getOid() begin with the prefix 'db' in order
  to be able to enforce the strict 3-tier application model.
  In the UI tier it should be forbidden to call an operation, whose name
  starts with the prefix db.
 * @author  Christoph Knabe
*/
public interface DbObject {

    public static final String oidSuffix = "_oid";

    /**Returns the OID of this object.*/
    public Integer getOid();

    /** Insert this object into its table in the database.
      Before inserting the object gets a new OID.
      @throws DuplicateInsertExc The actual object was already in the database
    */
    public void dbInsert() throws DuplicateInsertExc;

    /** Retrieves this object from its table in the database.
      Guarantees, that it has not been modified between the last select
      operation and this retrieve.
      You should in a transaction firstly call
      this method for each object participating in the transaction.
      @throws ConcurrentUpdateExc When retrieving this object from
        the database, a modification of the corresponding object in the
        database after the last select operation was detected.
    */
    public void dbRetrieve() throws ConcurrentUpdateExc;

    /** Update the representation in the database from this object.
     * @return number of objects actually updated
    */
    public int dbUpdate();

    /** Delete the representation of this object in the database.
     * Additonally mark this object as not valid by giving it the OID _noOid.
     * @return number of objects actually deleted
     */
    public int dbDelete();

    /** Returns the object with the OID i_oid from the database table i_table.
     *  The returned object will be dynamically an instance of the type given
     *  as first parameter to the constructor of i_table.
     *  Note1: The selected object should
     *  have no components of InputStream, Blob or Clob type, because
     *  their data will be not available after closing the Cursor used to get
     *  it.
     *  Note: Does not make use of the attribute values of the actual (dummy) object.
     *
     *  @throws OidNotFoundExc If in i_table no object with OID i_oid was found.
     */
    public /*static*/ DbObject dbSelectByOid(final jora.Table i_table,final Integer i_oid) throws OidNotFoundExc;

    /** Returns the one unique object by a "query by example"
     *  from the database.
     *  The returned object will be dynamically an instance of the same type
     *  as the actual example object.
     *  Note: The selected object should
     *  have no components of InputStream, Blob or Clob type, because
     *  their data will be not available after closing the Cursor used to get
     *  it.
     *
     *  @throws NotFoundExc If no object matching this example object was found.
     *  @throws NotUniqueExc If more than one object matches this example object.
     */
    public DbObject dbSelectUniqueByExample() throws ExampleNotFoundExc, ExampleNotUniqueExc;

    /**Returns the DbObject of the the class i_partnerClass, which is linked
      to from the actual object. The name of the class i_partnerClass is used
      to determine a) the Table, in which to serach, and b) the value of the
      OID looked for is taken from the own attribute, whose name is formed
      as a concatenation of the partner class name, and the suffix oidSuffix.
    */
    public DbObject dbGetlink(final Class i_partnerClass)
    throws NoPersistentClassExc, LinkFailure;

    /**Returns true, if in the actual table there is a pair with the oid-s
      of the DbObject-s i_left and i_right as foreign keys
    */
    public boolean dbAreLinked(final DbObjectImpl i_left,final DbObjectImpl i_right);


    public class ConcurrentUpdateExc extends multex.Exc {
      public ConcurrentUpdateExc(final String i_tableName, final Integer i_oid){
        super("{0}-object with OID {1} has been modified by another user. "
          + " You should repeat the transaction", i_tableName, i_oid
        );
      }
    }

    public class DuplicateInsertExc extends multex.Exc {
      public DuplicateInsertExc(final String i_tableName, final Integer i_oid){
        super("Cannot insert {0}-object with OID {1}, "
          + "as it is already in the database", i_tableName, i_oid
        );
      }
    }

    public static class ExampleNotFoundExc extends multex.Exc {
        public ExampleNotFoundExc(final jora.Table i_table, final String i_sql){
            super("Object not found by example in table \"{0}\", 1=sql",
                i_table.name(), i_sql
            );
        }
    }

    public static class ExampleNotUniqueExc extends multex.Exc {
        public ExampleNotUniqueExc(final jora.Table i_table, final String i_sql){
            super("Object selection by example not unique in table \"{0}\", 1=sql",
                i_table.name(), i_sql
            );
        }
    }

    public static class LinkFailure extends multex.Failure {
        public LinkFailure(final Exception i_cause, final String i_ownClassName, final String i_partnerOidFieldName, final String i_partnerTableName, final Integer i_partnerOid){
            super("Coming from class \"{0}\" with field \"{1}\" cannot find in partner table \"{2}\" the linked Object with OID {3} ",
                i_cause, i_ownClassName, i_partnerOidFieldName, i_partnerTableName, i_partnerOid
            );
        }
    }

    public static class NoPersistentClassExc extends multex.Exc {
        public NoPersistentClassExc(final Class i_class, final Class i_rootClass){
            super("Class {0} used for persistency is not a subclass of {1}",
                i_class.getName(), i_rootClass.getName()
            );
        }
    }

    public static class OidNotFoundExc extends multex.Exc {
        public OidNotFoundExc(final Integer i_oid,final jora.Table i_table){
            super("Object with OID {0} not found in table \"{1}\"",
                i_oid, i_table.name()
            );
        }
    }

    /** Get the JORA table responsible for storing objects of the same class
      as this.
      It is necessary to redefine this method in each concrete class of the
      application logic tier by <PRE>
      *   jora.Table dbTable(){return table;}
      </PRE>
    */
    public abstract jora.Table dbTable();

    public static final Integer noOid = null;

}