/*
 * DbObject.java
 * 2002-04-30  Knabe  dbRetrieve() checks new db__updateVersion counter.
 * 2002-04-11  Knabe  All operation names except getOid() begin with prefix db
 * 2001-05-28  Knabe  Preventing duplicate insert.
 *   All mutator operations only as protected, against usage from UI tier.
 * 2001-05-03  Knabe  OID type changed from int to Integer
 * 2001-04-22  Knabe  All object related DB operations here as protected
 * Created on 28. Februar 2001, 15:31
 */

package bank3tier.db;

/**
 *
 * @author  Christoph Knabe
 * @version
 */
/**Base class for all persistent objects. This class is purely generic.
  All application specific code (especially table creation) is collected
  in a subclass of {@link DbApplication}.
  Each object is identified by an OID. The OID (Object IDentification) is a
  number unique for each object in our application.
  All operation names except the getOid() begin with the prefix 'db' in order
  to be able to enforce the strict 3-tier application model.
   In the UI tier it should be forbidden to call an operation, whose name
  starts with the prefix db.
*/
public abstract class DbObjectImpl implements DbObject {

    /**The Class of this class*/
    private static Class _myClass = DbObjectImpl.class;

    public final Integer getOid(){return oid;}


   public void dbInsert() throws DuplicateInsertExc
    {
      if(getOid()!=noOid){throw new DuplicateInsertExc(dbTable().name(), getOid());}
      oid = _getNewOid();    db__updateVersion++;   dbTable().insert(this);
    }

   public void dbRetrieve() throws ConcurrentUpdateExc
    {
        final jora.Table table = dbTable();
        try{
            final DbObjectImpl dbObject = (DbObjectImpl)dbSelectByOid(table, getOid());
            //old approach: table.copy(dbObject, this);
            //new approach:
            if(dbObject.db__updateVersion!=this.db__updateVersion){ //meanwhile updated in the database
                throw new ConcurrentUpdateExc(table.name(), getOid());
            }
        } catch(ConcurrentUpdateExc ex){throw ex;
        } catch(Exception ex){
            throw new multex.Failure(
                "Failure retrieving object with OID {0} from table {1}",
                ex, getOid(), table.name()
            );
        }
    }

    public final int dbUpdate()
    { db__updateVersion++; return dbTable().update(this); }

    public final int dbDelete()
    { final int rowCount = dbTable().delete(this); oid = noOid; return rowCount;}

    public /*static*/ DbObject dbSelectByOid(final jora.Table i_table,final Integer i_oid) throws OidNotFoundExc
    {
        final jora.Cursor c = i_table.selectAll("where oid="+i_oid);
        final Object result = c.next();
        if(result==null){
            throw new OidNotFoundExc(i_oid, i_table);
        }
        //???Wegen Verlust der BLOBS usw. so sinnvoll?
        c.close();
        return (DbObject)result;
    }

    public DbObject dbSelectUniqueByExample() throws ExampleNotFoundExc, ExampleNotUniqueExc
    {
        final jora.Table table = dbTable();
        final jora.Cursor cursor = table.queryByExample(this);
        final Object[] allObjects = cursor.toArray();
        final String query = cursor.getQuery(); //Query only known after Cursor.next()
        if(allObjects.length>1){throw new ExampleNotUniqueExc(table, query);
        }else if(allObjects.length<1){
            throw new ExampleNotFoundExc(table, query); //???Parameter
        }
        return (DbObject)allObjects[0];
    }

    public DbObject dbGetlink(final Class i_partnerClass)
    throws NoPersistentClassExc, LinkFailure
    {
        dbCheckPersistentClass(i_partnerClass);
        final Class ownClass = getClass();
        String partnerTableName = "";
        String partnerOidFieldName = "";
        Integer partnerOid = null;
        try {
            //Get the Table of i_partnerClass:
            final java.lang.reflect.Field partnerTableField
            = i_partnerClass.getDeclaredField("table");
            partnerTableField.setAccessible(true);
            final Object partnerTableObject = partnerTableField.get(null);
            final jora.Table partnerTable = (jora.Table)partnerTableObject;
            //Determine the OID of the linked partner object:
            partnerTableName = partnerTable.name();
            partnerOidFieldName = partnerTableName.toLowerCase() + oidSuffix;
            final java.lang.reflect.Field partnerOidField
            = ownClass.getDeclaredField(partnerOidFieldName);
            partnerOidField.setAccessible(true);
            final Object partnerOidObject = partnerOidField.get(this);
            partnerOid = (Integer)partnerOidObject;
            //Get the object of the partner table with the partner OID:
            return dbSelectByOid(partnerTable, partnerOid);
        } catch(Exception ex){
            throw new LinkFailure(ex, ownClass.getName(), partnerOidFieldName, partnerTableName, partnerOid);
        }
    }

    public boolean dbAreLinked(final DbObjectImpl i_left,final DbObjectImpl i_right) {
        final Object[] links = dbTable().select(
            "WHERE " + i_left.dbTable().name() + oidSuffix + " = " + i_left.getOid()
            + " AND "+ i_right.dbTable().name() + oidSuffix + " = "+i_right.getOid()
        ).toArray();
        return links.length >= 1;
    }

    /**Checks, if the specified Class is a subclass of this class
      @throws NoPersistentClassExc If the specified Class is not a subclass of DbObjectImpl
    */
    /*package*/ static void dbCheckPersistentClass(final Class i_persistentClass)
    throws NoPersistentClassExc
    {
        if(!_myClass.isAssignableFrom(i_persistentClass)){
            throw new NoPersistentClassExc(i_persistentClass, _myClass);
        }
    }

    /**Unique object Identification for each DbObject instance*/
    private Integer        oid = noOid;

    /**Modification version counter for this object, shall be incremented
      by any insert oder update operation.
    */
    private int db__updateVersion = 0;

    private static int _lastOid = 0;

    /** Returns an object identification number unique for each object in our
      * application. In a non-demo application with real persistency the
      * _lastOid should be stored in the database, too, rather than in a
      * variable.
    */
    private static Integer _getNewOid(){return new Integer(++_lastOid);}


}